#include <cassert>
#include <cstring>
#include <cctype>
#include <stdio.h>      //!!

#include <string.h>

#include <windows.h>

#include "idx.h"


FileType detect_file_type(const u8* buf, const int size)
{
    assert(buf != 0);
    if(size < 4)        // i need a little more to work !
        return FT_NON_TEXT;                                      // <- verhalten ?

    const unsigned long  first4b = *(unsigned long*)buf;   // first 4 bytes from buffer
    const unsigned long  second4b = *(unsigned long*)(buf+4);   // second 4 bytes from buffer
	const unsigned short first2b = *(unsigned short*)buf;   // first 2 bytes fron buffer

    /*
     *  first some quick tests with "magic numbers"
     */

    if (first4b == 0x04034b50)      // .zip - "PK\003\004"
        return FT_ZIP;

    if (first4b == 0x21726152)      // .rar - "RAR!"
        return FT_RAR;

    if (((first2b&0xfeff) == 0xfaff)||      // MPEG1 L3 header
        ((first2b&0xf8ff) == 0xf0ff)||      // MPEG2 L3 header
        ((first4b&0x00ffffff) == 0x334449)) // .mp3 - "ID3" (ID3v2-tag)
        return FT_MP3;

    //  some filetypes to skip
       // pictures
    if ((first2b == 0xd8ff)||       // .jpg - JPEG
        (first4b == 0x38464947)||   // .gif - "GIF8"
        (first4b == 0x474e5089)||   // .png - "\x89PNG"
        (first2b == 0x4d42)||       // .bmp - "BM"
        // videos
		second4b == 0x7461646d ||	// .mov - "mdat"
		second4b == 0x766f6f6d ||	// .mov - "moov"
		first4b == 0x46464952||   // .avi | .wav | .mid - "RIFF" (little endian)
		first4b == 0x58464952||   // - " -              - "RIFX" (big endian)
		first4b == 0xb3010000||   // MPEG video  stream data
		first4b == 0xba010000||   // MPEG system stream data
		first4b == 0x5367674f||   // Ogg Data - "OggS"
		first4b == 0x75b22630)    // .asf (M$ "advanced" streaming format)
        return FT_NON_TEXT;

    /*
     *  no match ? ok ... let's look a little bit deeper !
     *  look for unique strings and special treatment for some filetypes
     */

    // make sure our buffer is big enough

    if(size< 12)
        return FT_NON_TEXT;

    // PDF + PostScript

const char* p = (const char*)buf;

    if( !strncmp( p, "%!PS-Adobe-", 10)||         // PostScript document
        !strncmp( p, "\004%!PS-Adobe-", 11)||     // PostScript document
        !strncmp( p, "\033%-12345X@PJL", 12)||    // HP Printer Job Language
        !strncmp( p, "%PDF-", 5) )                // PDF document
            return FT_PDF;

    // SGML ( html, xml, ...)

    int bsize = size;
    while(isspace(*buf) && (bsize--))      // skip leading white spaces
        { *buf++; };
    if(!bsize)
        return FT_POSSIBLE_TEXT;

p = (const char*)buf;

    if( !strnicmp(p, "<html", 5)||
        !strnicmp(p, "<head", 5)||
        !strnicmp(p, "<body", 5)||
        !strnicmp(p, "<title", 6)||
        !strnicmp(p, "<!doctype", 9)||
        !strnicmp(p, "<?xml", 5)||
        !strnicmp(p, "<!--", 4)||
        !strnicmp(p, "<!subdoc", 8) )
            return FT_HTML;
            
    // can't handle it :-(

    return FT_POSSIBLE_TEXT;
}

// ----------------------------------------------------------------------------


static const u32 is_alpha_tbl[256/32] =
{
	0x00000000,	0x00000000,
	0x07fffffe,	0x07fffffe,
	0x00000000,	0x00000000,
	0x90400010,	0x10400010
};

static inline bool _is_alpha(const int c)
{
	return (is_alpha_tbl[c >> 5] & (1ul << (c & 0x1f))) != 0;
}



bool contains_text(const u8* buf, const int size, int* start_ofs, int* end_ofs)
{
    assert(buf != 0);

    // initialize variables

    int bytes_left = size;

    int first = 0;      // first occurence of text
    int last = 0;       // last ...

    int alpha_count=0; int ws_count=0; int word_count=0;    // some counters
    int max_word_len=0; int max_textblock_len=0; int text_len = 0;
    int word_len = 0; int textblock_len = 0;

    int mode = 1;  // 1: search for alpha; 2: walk word; 3: ws_hit, walk textblock

    // process buffer

    while (bytes_left--)
    {
        const unsigned int c = *buf++;
        const bool is_alpha = _is_alpha(c);
        const int is_space = isspace(c);

        if(is_alpha)
        {
            last = size - bytes_left;   // set end_ofs
            if(!first) first = last;    // set start_ofs

            mode = 2;
            alpha_count++;
            word_len++;
            textblock_len++;
        }
        else if(is_space)
        {
            if(mode != 1)     // count only ws inside of text
            {
                if(mode == 2)
                {
                    if(max_word_len < word_len) max_word_len = word_len;
                    if(word_len > 3) word_count++;      // count only then long ones
                    word_len = 0;
                }
                mode = 3;
                ws_count++;
                textblock_len++;
            }
        }
        else if(mode != 1)
        {
            mode = 1;
            if(max_word_len < word_len) max_word_len = word_len;
            if(max_textblock_len < textblock_len) max_textblock_len = textblock_len;
            if(word_len > 3)    // count only then long ones
            {
                word_count++;
                text_len += textblock_len;
            }
            word_len = 0;
            textblock_len = 0;
        }
    }

    if(max_word_len < word_len) max_word_len = word_len;
    if(max_textblock_len < textblock_len) max_textblock_len = textblock_len;

    *start_ofs = word_count;//first;
    *end_ofs = text_len;//last;

    char txt[1000];
    sprintf(txt, "alpha_count: %d \nws_count: %d \nword_count: %d \n\nmax_word_len: %d \nmax_textblock_len: %d \ntext_len: %d \nsize: %d\n",
        alpha_count, ws_count, word_count, max_word_len, max_textblock_len, text_len, size);
    OutputDebugString( txt );
    sprintf(txt, "alpha/size: %.2f , ws/size: %.2f , text/size: %.2f , ws/text: %.2f",
        100.f*alpha_count/size, 100.f*ws_count/size, 100.f*text_len/size, (text_len>0)?100.f*ws_count/text_len:0.f);
    OutputDebugString( txt );

    // text or not ? :)

    return true;   // <- change !!!!!
}



