// JW

#include <cassert>
#include <stdio.h>

#include <vector>
#include <algorithm>

#include <windows.h>
#include <process.h>	// _beginthreadex

#include "misc.h"

#include "exthashtbl.h"
#include "idx.h"

u64 total_text = 0, total_words = 0, unique_words = 0;


//////////////////////////////////////////////////////////////////////////////
//
// memory management
// pool allocator, reset every index create/read
//
//////////////////////////////////////////////////////////////////////////////

// normal mem and index mapping pools
static struct Pool
{
	u8* ptr;
	u32 end;
	u32 allocated;
	u32 size;
	u8* freelist;	// freed buckets in index mapping
					// each bucket in list starts with pointer to next bucket
}
pools[2];
static Pool& idx_pool = pools[1];


// index mapping

static const char* idx_filename = "idx.dat";
static HANDLE hFile, hMap;

static int idx_create_file()
{
	idx_pool.size = 64*1024*1024;

	hFile = CreateFile(idx_filename, GENERIC_READ|GENERIC_WRITE, 0, 0, CREATE_ALWAYS, FILE_FLAG_SEQUENTIAL_SCAN, 0);
	if(hFile == INVALID_HANDLE_VALUE)
		return -1;
	hMap = CreateFileMapping(hFile, 0, PAGE_READWRITE, 0, idx_pool.size, 0);
	idx_pool.ptr = (u8*)MapViewOfFile(hMap, FILE_MAP_ALL_ACCESS, 0, 0, idx_pool.size);

	return 0;
}


static u8* idx_map_file()
{
	hFile = CreateFile(idx_filename, GENERIC_READ, 0, 0, OPEN_EXISTING, FILE_FLAG_RANDOM_ACCESS, 0);
	if(hFile == INVALID_HANDLE_VALUE)
		return 0;
	hMap = CreateFileMapping(hFile, 0, PAGE_READONLY, 0, 0, 0);
	idx_pool.ptr = (u8*)MapViewOfFile(hMap, FILE_MAP_READ, 0, 0, 0);
	idx_pool.end = GetFileSize(hFile, 0);

	return idx_pool.ptr;
}


static void idx_close_file()
{
	// file wasn't mapped
	if(!idx_pool.ptr)
		return;

	// flush changes; also can't set file ptr with file mapped
	UnmapViewOfFile(idx_pool.ptr);
	CloseHandle(hMap);

	// truncate file to actual length
	SetFilePointer(hFile, idx_pool.end, 0, FILE_BEGIN);
	SetEndOfFile(hFile);
	CloseHandle(hFile);

	idx_pool.ptr = 0;	// will need to re-map at first alloc after this
}


// used by ExtHashTbl
u8* mem_alloc(u32 size, bool in_idx)
{
	Pool& p = pools[in_idx];

	// memory not yet allocated
	if(!p.ptr)
	{
		assert(!in_idx && "mem_alloc for idx, without file mapped");
		p.ptr = (u8*)VirtualAlloc(0, p.size = 8*1024*1024, MEM_COMMIT, PAGE_READWRITE);
	}

	u8* ptr;

	// allocating buckets in index mapping
	if(in_idx && size == BUCKET_SIZE)
	{

		// get from freelist
		ptr = p.freelist;
		if(ptr)
		{
			p.freelist = *(u8**)ptr;
			p.allocated += size;
			return ptr;
		}
  
		// alloc from pool - make sure they're aligned
		p.end = round_up(p.end, BUCKET_SIZE);
	}

	if(p.end >= p.size)
		return 0;

	ptr = p.ptr + p.end;
	p.end += size;
	p.allocated += size;

	return ptr;
}


static void mem_free(u8* ptr, u32 size)
{
	// what pool contains ptr?
	bool in_idx = false;
	Pool* p = pools;
	// .. not normal mem
	if((unsigned)(ptr - p->ptr) >= p->end)
	{
		in_idx = true;
		p++;
	}
	// .. nor index mapping
	if((unsigned)(ptr - p->ptr) >= p->end)
	{
		assert(!"invalid pointer to mem_free (not in either pool)");
		return;
	}

	p->allocated -= size;

	// it's at the end of the pool - remove from end
	if(ptr+size == p->ptr+p->end)
		p->end -= size;
	// need to add to freelist
	else
	{
		assert(in_idx && size == BUCKET_SIZE &&
			"mem_free: not at end of pool, nor bucket in index mapping");
		*(u8**)ptr = p->freelist;
		p->freelist = ptr;
	}
}


// release all memory used by previous index
static void mem_reset()
{
	for(int i = 0; i < 2; i++)
	{
		Pool* p = &pools[i];
		p->allocated = 0;
		p->end = 0;
		p->freelist = 0;
	}
}


// offset -> pointer
static u8* idx_ptr(u32 ofs)
{
	assert(ofs < idx_pool.end && "invalid offset in index mapping");

	return idx_pool.ptr + ofs;
}


// pointer -> offset
static u32 idx_ofs(u8* ptr)
{
	u32 ofs = ptr - idx_pool.ptr;
	assert(ofs < idx_pool.end && "invalid pointer in index mapping");

	return ofs;
}



//////////////////////////////////////////////////////////////////////////////
//
// variable length list of files where a word occurs; includes frequency
//   and block number of first occurrence.
//
// data structure: must allow O(1) insert and saving to file (=> relocatable)
//   list of buckets (space for several elements + pointer to next);
//   each instance has a ptr to the first bucket and last element
//   (buckets are naturally aligned => we can check if full without counter)
//
//////////////////////////////////////////////////////////////////////////////

// Occurrence field widths (not needed by other modules)

const int FREQ_BITS = 6;
const int MAX_FREQ = (1 << FREQ_BITS) - 1;

const int SEARCH_BITS = 2;
const int MAX_SEARCHES = (1 << SEARCH_BITS) - 1;

// 32 bit
struct Occurrence
{
	u32 freq  : FREQ_BITS;		// clamped; used for best match
	u32 word  : SEARCH_BITS;	// this is a hit for which search word?
								// filled in by dump_occs
	u32 block : BLOCK_BITS;		// first occ.'s block # in file (clamped) 
	u32 fn    : FILE_NUM_BITS;
};



// "fits in bucket" test requires each instance to be naturally aligned,
// and assumes total size = 2^n and next field comes immediately after array
struct Bucket
{
	Occurrence o[BUCKET_ENTRIES];
	u32 next_ofs;		// offset of next bucket in list
};

static int seq;


bool care;

// return value: was it actually added?
static bool add_occ(Word* w)
{
seq++;
	// BCB is too stupid to see these aren't used if goto taken
	Occurrence* o;
	int last_b_entry;

	// no buckets allocated yet
	if(!w->buckets)
		goto alloc;

	// already in list (at end)?
	o = (Occurrence*)idx_ptr(w->last);
	if(o->fn == cur_fn)
		goto skip_insert;

	// needs to be added to list
	assert(cur_fn >= o->fn);

	// no more room left in this bucket - alloc another
	last_b_entry = (w->last % sizeof(Bucket))/4 + 1;
	if(last_b_entry == BUCKET_ENTRIES)
	{
		// don't want to alloc any more buckets - make it a stopword
		if(w->buckets*BUCKET_ENTRIES >= MAX_OCCS)
		{
			if(w->key == 1554008311)
				w=w;
			// free buckets allocated for this word
			u32 ofs = w->first*sizeof(Bucket);
			for(unsigned i = 0; i < w->buckets; i++)
			{
				u8* p = idx_ptr(ofs);
				assert(ofs != 0);
				mem_free(p, sizeof(Bucket));
				ofs = ((Bucket*)p)->next_ofs;
			}

			w->is_stopword = true;
			return false;
		}

alloc:
		u32 b_ofs = idx_ofs(mem_alloc(sizeof(Bucket), true));
		assert(b_ofs % sizeof(Bucket) == 0);

		// first time: point to start of list
		if(!w->buckets)
			w->first = b_ofs / sizeof(Bucket);
		// attach to previous bucket
		else
			*(u32*)(o+1) = b_ofs;

		w->last = b_ofs;
		w->buckets++;
		o = (Occurrence*)idx_ptr(w->last);
	}
	// fits in current bucket
	else
	{
		w->last += 4;
		o++;
	}

	o->fn = cur_fn;
	o->block = cur_block;
	o->freq = 0;

skip_insert:

	// update freq
	if(o->freq < MAX_FREQ)
		o->freq++;

	return true;
}




static int num_search_words;
static char search_words[MAX_SEARCHES][MAX_WORD_LEN+1];
static char* cur_search_word;
static bool found_word;

static std::vector<Occurrence> results;
static std::vector<Occurrence>::iterator res_it;

static u8 rank[MAX_FILES];


static int dump_occs(Word* w)
{
	if(!w || w->is_stopword)
		return -1;

	u32* p = (u32*)idx_ptr(w->first*sizeof(Bucket));
	u32* last = (u32*)idx_ptr(w->last);

	int buckets_left = w->buckets;
	while(buckets_left--)
	{
		for(int i = 0; i < BUCKET_ENTRIES; i++)
		{
			Occurrence o = *(Occurrence*)p;

			unsigned r = rank[o.fn];
			unsigned hits = r >> 6;
			unsigned total_freq = min((r & 0x3f) + o.freq, 0x3f);
			rank[o.fn] = ((hits+1) << 6) | total_freq;
//printf("fn %d, its rank: %d\n", o.fn, rank[o.fn]);
			o.word = num_search_words;	// associate with search word

			results.push_back(o);


			if(p++ == last)
			{
				assert(buckets_left == 0);
				return 0;	// *p might not be valid - don't just break
			}
		}

		// jump to next bucket
		p = (u32*)idx_ptr(*p);
	}

	return 0;
}


// sort a vector containing Occurrences of all search words.
// best match order means files containing the most searched for words
// come first; 2nd criteria is total count of all the search words.
static bool Occ_less(const Occurrence& o1, const Occurrence& o2)
{
	u32 fn1 = o1.fn, fn2 = o2.fn;

	// # of search words that occur in a file + total freq
	u8 r1 = rank[fn1], r2 = rank[fn2];
	if(r1 != r2)
		return r1 > r2;

	if(fn1 != fn2)
		return fn1 < fn2;

        return o1.word < o2.word;
}



//////////////////////////////////////////////////////////////////////////////
//
// worker functions for index
//
//////////////////////////////////////////////////////////////////////////////

/* struct Word */

static ExtHashTbl/*<Word>*/ ext_hash_tbl;


// FNV1-A
// not really time critical - ~1% CPU usage.
// could improve by unrolling 8x with MMX, though.
/* static */u32 hash(const char* str, int len)
{
	u32 h;

	while(len--)
	{
		h ^= *(const u8*)str++;
		h *= 0x01000193;
	}

	return h;
}


// when inserting words, copy from this template
// can set member is_stopword to have it become a stopword directly
static Word new_w;


// what to do with each word we find?
static enum
{
	INSERT,		// normal operation: add word to index

	GET_OCCS,	// we are being called for each search word;
				// append its occurrence list to results

	COMPARE		// set found_word if word == cur_search_word
				// (used to get surrounding text)
}
word_action;


// called by scanners for each word found. action taken depends on mode above.
// in all cases, we get the word stem, hash, and look up the Word entry.
//
// in GET_OCCS and COMPARE modes, we were called by extract_words.
// return -1 to have it stop scanning (if too many search words / word found,
// respectively). otherwise, return value is ignored.
//
// note: word string is modified in-place by stemmer.
int lookup_word(char* word, int len)
{
	if(len < MIN_WORD_LEN)
		return 0;

#ifdef DUMP_AR
static FILE* fa;
if(!fa)
{
	fa=fopen("d:\\projects\\info4\\idx\\accept.csv", "w");
	setvbuf(fa, 0, _IONBF, 0);
}
extern float score,bias;
#endif


	get_word_stem(word, len);
	word[len] = 0;	// 0-terminate

	// we're looking for a word for context
	if(word_action == COMPARE)
	{
		// quit extracting if word matches, otherwise continue
		if(!strcmp(word, cur_search_word))
		{
			found_word = true;
			return -1;
		}
		return 0;
	}

	u32 h = hash(word, len);
	Word* w = ext_hash_tbl.search(h);

	// append w's occurrence list to results
	if(word_action == GET_OCCS)
	{
		if(num_search_words >= MAX_SEARCHES)
			return -1;
		strcpy(&search_words[num_search_words][0], word);

		dump_occs(w);
		num_search_words++;		// after dump_occs!

		return 0;
	}

	// not found
	if(!w)
	{
#ifdef DUMP_AR
fprintf(fa, "%.2f, %.2f, %s\n", score, bias, word);
#endif
		unique_words++;

		new_w.key = h;
		w = ext_hash_tbl.insert(new_w);	// passed by reference
	}

	total_text += len;
	total_words++;

	if(!w->is_stopword)
		add_occ(w);

	return 0;
}


// index file header
static struct IdxHeader
{
	u8* ptr;				// address of index mapping when file was created
							// used to fixup ext_hash_tbl's pointers

	ExtHashTbl/*<Word>*/ ext_hash_tbl;

	char path[MAX_PATH+1];	// the path that was indexed
}
* header;


static HANDLE hThread = INVALID_HANDLE_VALUE;
static unsigned idx_thread;

// set by thread when done; used by idx_status
static bool idx_succeeded;


// do the actual indexing in the background.
// done <==> idx_status != IS_ACTIVE.
// we generate the file list here, because it may take a while;
// frontend will have to deal with increasing total_bytes (and check for 0!).
// num_files > 0 indicates the file list is ready.
static unsigned __stdcall idx_thread_func(void* param)
{
	word_action = INSERT;

	const char* path = (const char*)param;
	if(gen_file_list(path) == 0)	// succeeded
	{
		idx_succeeded = read_files_in_list() == 0;
		header->ext_hash_tbl = ext_hash_tbl;

		idx_close_file();
	}

	_endthreadex(0);
	return 0;
}


// 'scan' a file containing stopwords; these will be inserted as such,
// because we set the bit directly in the new_w template.
// this is not the complete stopword list, which is made dynamically;
// we pre-insert common words (that would likely become stopwords)
// to cut down on alloc/free in the index, reducing fragmentation.
static void insert_stopwords()
{
// TODO: move into read
	HANDLE hFile = CreateFile("stopwords.txt", GENERIC_READ, FILE_SHARE_READ, 0, OPEN_EXISTING, FILE_FLAG_RANDOM_ACCESS, 0);
	if(hFile == INVALID_HANDLE_VALUE)
		return;

	const int BUF_SIZE = 32*1024;
	u8* buf = mem_alloc(BUF_SIZE, false);

	u32 bytes_read;
	ReadFile(hFile, buf, BUF_SIZE, &bytes_read, 0);

	new_w.is_stopword = 1;
	extract_words(buf, bytes_read);
	new_w.is_stopword = 0;

	mem_free(buf, BUF_SIZE);
	CloseHandle(hFile);
}



//////////////////////////////////////////////////////////////////////////////
//
// interface to frontend
//
//////////////////////////////////////////////////////////////////////////////


IdxStatus idx_status()
{
	if(hThread != INVALID_HANDLE_VALUE)
	{
		if(WaitForSingleObject(hThread, 0) == WAIT_TIMEOUT)
			return IS_ACTIVE;

		// just completed
		CloseHandle(hThread);
		hThread = INVALID_HANDLE_VALUE;
	}

	return idx_succeeded? IS_SUCCEEDED : IS_FAILED;
}


// discard previous index, and (re)load from file
// return -1 on error (including "index in progress")
//
// note: may take a few seconds, especially if getting file list from CD
// (normal HD mode is faster, since Windows caches the FS)
int read_idx()
{
	// already running
	if(idx_status() == IS_ACTIVE)
	{
		assert(!"read_idx while indexing active");
		return -1;
	}

	header = (IdxHeader*)idx_map_file();
	if(!header)
		return -1;

	ext_hash_tbl = header->ext_hash_tbl;
	long delta = (u8*)header - header->ptr;	// delta between old and new pointers
	ext_hash_tbl.rebase(delta);

	return gen_file_list(cur_path = header->path);
}


// start indexing all files in path, including subdirectories.
// asynchronous - progress is indicated by total_bytes_read etc.;
// idx_status changes from IS_ACTIVE to IS_SUCCEEDED or IS_FAILED
// when indexing is complete (including errors and cancels).
// return -1 on error (already indexing, or cannot map index file),
// or 0 for success.
long create_idx(const char* path)
{
// TODO: not completely thread safe (2 concurrent calls => boom)
	// already running
	if(idx_status() == IS_ACTIVE)
	{
		assert(!"create_idx while indexing active");
		return -1;
	}

	// free all memory allocated for previous index.
	mem_reset();

	// reset stats
	total_text = total_words = unique_words = 0;

	idx_close_file();
	long error = idx_create_file();
	if(error < 0)
		return error;

	header = (IdxHeader*)mem_alloc(sizeof(IdxHeader), true);
	header->ptr = (u8*)header;
	strncpy(header->path, path, MAX_PATH);

	ext_hash_tbl.init();
	insert_stopwords();

	// start indexing thread
	hThread = (HANDLE)_beginthreadex(0, 0, idx_thread_func, (void*)path, 0, &idx_thread);

	return idx_pool.allocated;	// size of index
}


// request asynchronous index creation to be cancelled.
// indexing will complete soon with idx_status == IS_FAILED;
// the index data is not destroyed, and may be used in a search.
//
// returns -1 if indexing not in progress, or 0 for success.
int cancel_idx()
{
	if(idx_status() != IS_ACTIVE)
	{
		assert("cancel_idx while not indexing");
		return -1;
	}

	// note:
	// the read code will abort after the current block is finished.
	// we could send an APC to kick it out of its wait,
	// but it completes fast enough anyway. don't bother.

	read_cancelled = true;

	return 0;
}


// search for all occurrences of the words in string in the index.
// if there are more than MAX_SEARCHES = 3 words, the rest are ignored, and
// the return value indicates the offset of the first ignored character.
// otherwise, the return value is 0 for success or -1 for failure.
//
// results are returned in best match order by get_next_search_result.
long search(const char* string)
{
	// reset search results
	num_search_words = 0;
	results.clear();
	memset(rank, 0, sizeof(rank));

	word_action = GET_OCCS;

	// for each word in the search string:
	//   append its list of Occurrences to results vector
	long ret = extract_words((const u8*)string, strlen(string)+1);
		// note: include trailing \0 as a terminator
		//       => don't need to flush output buffer

	// 'best match' sort (see above)
	std::sort(results.begin(), results.end(), Occ_less);	
	res_it = results.begin();

	return ret;
}


// retrieve results from the most recent search in best match order.
// return value is -1 if no (more) results available;
// otherwise, filename, freq, and context are filled, and 0 is returned.
//
// searching for surrounding text is slow, and can be disabled
// by passing context = 0. position in output array is reset
// when context was 0 last time, and non-null now
// (allows app to get list, then populate with surrounding text
// transparently; no explicit reset)
//
// freq is the sum of the frequences of all search words.
// if -freq > 0, it's a lower bound (since one of the freqs was clamped).
// filename and context receive pointers to static buffers,
//   which are overwritten by the next call to this function.
int get_next_search_result(const char*& filename, int& freq, const char** context)
{
	// reset; see above
	static const char** last_context;
	if(!last_context && context)
		res_it = results.begin();
	last_context = context;

	// no search started, or last result already returned
	if(!results.size() || res_it == results.end())
		return -1;

	Occurrence o = *res_it;
	int hits = rank[o.fn] >> 6;	// how many of the search words occur in this file
								// i.e. next hits-1 results are for this file

	filename = get_filename(o.fn);
	freq = 0;

	static char ctx_buf[256];
	int pos = 0;
	if(context)
	{
		*context = ctx_buf;
		memset(ctx_buf, 0, sizeof(ctx_buf));
			// don't return old data if search below fails
	}

	word_action = COMPARE;

	bool overflow = false;	// >= word's freq was clamped

	// for each search word that occurs in this file:
	//   get surrounding text; add total # occurrences to freq total
	for(int h = 0; h < hits; h++)
	{
		o = *res_it++;

		if(!context)
		{
			if(o.freq == MAX_FREQ)
				overflow = true;
			freq += o.freq;
			continue;
		}

		// lookup_word in COMPARE mode will look for this word, stopping when found
		cur_search_word = &search_words[o.word][0];
		int csw_len = strlen(cur_search_word);

		found_word = false;

		const u8* buf;
		long block = o.block;
		long size, ofs;

		// read+scan blocks starting at o.block, until EOF reached or found
		for(;;)
		{
			ofs = read_block(o.fn, block++, buf, size);
			if(ofs == -1)
				break;		// EOF

			if(found_word)
			{
				// copy 10 chars before and after
				long start = max(ofs-csw_len-11, 0), end = min(ofs+9, size);
                        int len = end-start;
                        char* p = &ctx_buf[pos];
                        memcpy(p, &buf[start], len);
                        pos += len+5;
		do
			if(*p == '\r' || *p == '\n' || *p == 0)
				*p = ' ';
		while(len-- && p++);
		strcat(p, " ... ");
//				pos += sprintf(&ctx_buf[pos], "%.*s ... ", end-start, buf+start);

				if(o.freq == MAX_FREQ)
					overflow = true;
				freq += o.freq;

				break;
			}
		}
	}

	if(overflow) freq = -freq;	// notify frontend

	return 0;
}
