
/* This is the Porter stemming algorithm, coded up in ANSI C by the
   author. It may be be regarded as cononical, in that it follows the
   algorithm presented in

   Porter, 1980, An algorithm for suffix stripping, Program, Vol. 14,
   no. 3, pp 130-137,

   only differing from it at the points maked --DEPARTURE-- below.

   See also http://www.tartarus.org/~martin/PorterStemmer

   The algorithm as described in the paper could be exactly replicated
   by adjusting the points of DEPARTURE, but this is barely necessary,
   because (a) the points of DEPARTURE are definitely improvements, and
   (b) no encoding of the Porter stemmer I have seen is anything like
   as exact as this version, even with the points of DEPARTURE!

   You can compile it on Unix with 'gcc -O3 -o stem stem.c' after which
   'stem' takes a list of inputs and sends the stemmed equivalent to
   stdout.

   The algorithm as encoded here is particularly fast.

   Release 1
*/

#include <string.h>  /* for memmove */

#define TRUE 1
#define FALSE 0

/* The main part of the stemming algorithm starts here. b is a buffer
   holding a word to be stemmed. The letters are in b[k0], b[k0+1] ...
   ending at b[k]. In fact k0 = 0 in this demo program. k is readjusted
   downwards as the stemming progresses. Zero termination is not in fact
   used in the algorithm.

   Note that only lower case sequences are stemmed. Forcing to lower case
   should be done before stem(...) is called.
*/

static char * b;       /* buffer for word to be stemmed */
static int k,k0,j;     /* j is a general offset into the string */

/* cons(i) is TRUE <=> b[i] is a consonant. */

static int cons(int i)
{  switch (b[i])
   {  case 'A': case 'E': case 'I': case 'O': case 'U': return FALSE;
      case 'Y': return (i==k0) ? TRUE : !cons(i-1);
      default: return TRUE;
   }
}

/* m() measures the number of consonant sequences between k0 and j. if c is
   a consonant sequence and v a vowel sequence, and <..> indicates arbitrary
   presence,

      <c><v>       gives 0
      <c>vc<v>     gives 1
      <c>vcvc<v>   gives 2
      <c>vcvcvc<v> gives 3
      ....
*/

static int m()
{  int n = 0;
   int i = k0;
   while(TRUE)
   {  if (i > j) return n;
      if (! cons(i)) break; i++;
   }
   i++;
   while(TRUE)
   {  while(TRUE)
      {  if (i > j) return n;
            if (cons(i)) break;
            i++;
      }
      i++;
      n++;
      while(TRUE)
      {  if (i > j) return n;
         if (! cons(i)) break;
         i++;
      }
      i++;
   }
}

/* vowelinstem() is TRUE <=> k0,...j contains a vowel */

static int vowelinstem()
{  int i; for (i = k0; i <= j; i++) if (! cons(i)) return TRUE;
   return FALSE;
}

/* doublec(j) is TRUE <=> j,(j-1) contain a double consonant. */

static int doublec(int j)
{  if (j < k0+1) return FALSE;
   if (b[j] != b[j-1]) return FALSE;
   return cons(j);
}

/* cvc(i) is TRUE <=> i-2,i-1,i has the form consonant - vowel - consonant
   and also if the second c is not w,x or y. this is used when trying to
   restore an e at the end of a short word. e.g.

      cav(e), lov(e), hop(e), crim(e), but
      snow, box, tray.

*/

static int cvc(int i)
{  if (i < k0+2 || !cons(i) || cons(i-1) || !cons(i-2)) return FALSE;
   {  int ch = b[i];
      if (ch == 'W' || ch == 'X' || ch == 'Y') return FALSE;
   }
   return TRUE;
}

/* ends(s) is TRUE <=> k0,...k ends with the string s. */

static int ends(char * s)
{  int length = s[0];
   if (s[length] != b[k]) return FALSE; /* tiny speed-up */
   if (length > k-k0+1) return FALSE;
   if (memcmp(b+k-length+1,s+1,length) != 0) return FALSE;
   j = k-length;
   return TRUE;
}

/* setto(s) sets (j+1),...k to the characters in the string s, readjusting
   k. */

static void setto(char * s)
{  int length = s[0];
   memmove(b+j+1,s+1,length);
   k = j+length;
}

/* r(s) is used further down. */

static void r(char * s) { if (m() > 0) setto(s); }

/* step1ab() gets rid of plurals and -ed or -ing. e.g.

       caresses  ->  caress
       ponies    ->  poni
       ties      ->  ti
       caress    ->  caress
       cats      ->  cat

       feed      ->  feed
       agreed    ->  agree
       disabled  ->  disable

       matting   ->  mat
       mating    ->  mate
       meeting   ->  meet
       milling   ->  mill
       messing   ->  mess

       meetings  ->  meet

*/

static void step1ab()
{  if (b[k] == 'S')
	{ if (ends("\04" "SSES"))  k -= 2;  else
	  if (ends("\03" "IES"))  setto("\01" "I");  else
	  if (b[k-1] != 'S')  k--; 
   }
if (ends("\03" "EED")) { if (m() > 0)  k--;  } else
   if ((ends("\02" "ED") || ends("\03" "ING")) && vowelinstem())
   {  k = j;
	  if (ends("\02" "AT"))  setto("\03" "ATE");   else
	  if (ends("\02" "BL"))  setto("\03" "BLE");   else
	  if (ends("\02" "IZ"))  setto("\03" "IZE");   else
      if (doublec(k))
      {  k--;
		 {  int ch = b[k];
            if (ch == 'L' || ch == 'S' || ch == 'Z') k++;
         }
      }
      else if (m() == 1 && cvc(k)) setto("\01" "E");
   }
}

/* step1c() turns terminal y to i when there is another vowel in the stem. */

static void step1c() { if (ends("\01" "Y") && vowelinstem()) b[k] = 'I'; }


/* step2() maps double suffices to single ones. so -ization ( = -ize plus
   -ation) maps to -ize etc. note that the string before the suffix must give
   m() > 0. */

static void step2() { switch (b[k-1])
{
    case 'A': if (ends("\07" "ATIONAL")) { r("\03" "ATE"); break; }
              if (ends("\06" "TIONAL")) { r("\04" "TION"); break; }
              break;
    case 'C': if (ends("\04" "ENCI")) { r("\04" "ENCE"); break; }
              if (ends("\04" "ANCI")) { r("\04" "ANCE"); break; }
              break;
    case 'E': if (ends("\04" "IZER")) { r("\03" "IZE"); break; }
              break;
    case 'L': if (ends("\03" "BLI")) { r("\03" "BLE"); break; } /*-DEPARTURE-*/

 /* To match the published algorithm, replace this line with
    case 'l': if (ends("\04" "abli")) { r("\04" "able"); break; } */

              if (ends("\04" "ALLI")) { r("\02" "AL"); break; }
              if (ends("\05" "entli")) { r("\03" "ENT"); break; }
              if (ends("\03" "eli")) { r("\01" "E"); break; }
              if (ends("\05" "OUSLI")) { r("\03" "OUS"); break; }
              break;
    case 'O': if (ends("\07" "IZATION")) { r("\03" "IZE"); break; }
              if (ends("\05" "ATION")) { r("\03" "ATE"); break; }
              if (ends("\04" "ATE")) { r("\03" "ATE"); break; }
              break;
    case 'S': if (ends("\05" "ALISM")) { r("\02" "AL"); break; }
              if (ends("\07" "IVENESS")) { r("\03" "IVE"); break; }
              if (ends("\07" "FULNESS")) { r("\03" "FUL"); break; }
              if (ends("\07" "OUSNESS")) { r("\03" "OUS"); break; }
              break;
    case 'T': if (ends("\05" "SLITI")) { r("\02" "AL"); break; }
              if (ends("\05" "IVITI")) { r("\03" "IVE"); break; }
              if (ends("\06" "BILITI")) { r("\03" "BLE"); break; }
              break;
    case 'G': if (ends("\04" "LOGI")) { r("\03" "LOG"); break; } /*-DEPARTURE-*/

 /* To match the published algorithm, delete this line */

} }

/* step3() deals with -ic-, -full, -ness etc. similar strategy to step2. */

static void step3() { switch (b[k])
{
    case 'E': if (ends("\05" "ICATE")) { r("\02" "IC"); break; }
              if (ends("\05" "ATIVE")) { r("\00" ""); break; }
              if (ends("\05" "ALIZE")) { r("\02" "AL"); break; }
              break;
    case 'I': if (ends("\05" "ICITI")) { r("\02" "IC"); break; }
              break;
    case 'L': if (ends("\04" "ICAL")) { r("\02" "IC"); break; }
              if (ends("\03" "FUL")) { r("\00" ""); break; }
              break;
    case 'S': if (ends("\04" "NESS")) { r("\00" ""); break; }
              break;
} }

/* step4() takes off -ant, -ence etc., in context <c>vcvc<v>. */

static void step4()
{  switch (b[k-1])
    {  case 'A': if (ends("\02" "AL")) break; return;
       case 'C': if (ends("\04" "ANCE")) break;
                 if (ends("\04" "ENCE")) break; return;
       case 'E': if (ends("\02" "ER")) break; return;
       case 'I': if (ends("\02" "IC")) break; return;
       case 'L': if (ends("\04" "ABLE")) break;
                 if (ends("\04" "IBLE")) break; return;
       case 'N': if (ends("\03" "ANT")) break;
                 if (ends("\05" "EMENT")) break;
                 if (ends("\04" "MENT")) break;
                 if (ends("\03" "ENT")) break; return;
       case 'O': if (ends("\03" "ION") && (b[j] == 'S' || b[j] == 'T')) break;
                 if (ends("\02" "OU")) break; return;
                 /* takes care of -ous */
       case 'S': if (ends("\03" "ISM")) break; return;
       case 'T': if (ends("\03" "ATE")) break;
                 if (ends("\03" "ITI")) break; return;
       case 'U': if (ends("\03" "OUS")) break; return;
       case 'V': if (ends("\03" "IVE")) break; return;
       case 'Z': if (ends("\03" "IZE")) break; return;
       default: return;
    }
    if (m() > 1) k = j;
}

/* step5() removes a final -e if m() > 1, and changes -ll to -l if
   m() > 1. */

static void step5()
{  j = k;
   if (b[k] == 'e')
   {  int a = m();
      if (a > 1 || a == 1 && !cvc(k-1)) k--;
   }
   if (b[k] == 'l' && doublec(k) && m() > 1) k--;
}

/* In stem(p,i,j), p is a char pointer, and the string to be stemmed is from
   p[i] to p[j] inclusive. Typically i is zero and j is the offset to the last
   character of a string, (p[j+1] == '\0'). The stemmer adjusts the
   characters p[i] ... p[j] and returns the new end-point of the string, k.
   Stemming never increases word length, so i <= k <= j. To turn the stemmer
   into a module, declare 'stem' as extern, and delete the remainder of this
   file.
*/

int stem(char * p, int i, int j)
{  b = p; k = j; k0 = i; /* copy the parameters into statics */
   //if (k <= k0+1) return k; /*-DEPARTURE-*/

   /* With this line, strings of length 1 or 2 don't go through the
      stemming process, although no mention is made of this in the
      published algorithm. Remove the line to match the published
      algorithm. */

   step1ab(); step1c(); step2(); step3(); step4(); step5();
   return k;
}

/*--------------------stemmer definition ends here------------------------*/

#include <stdio.h>

//#include <ctype.h>
       /* for isupper, islower, tolower */
#include "idx.h"

static char * s;         /* a char * (=string) pointer; passed into b above */

void get_word_stem(char* word, int& len){
  char* s = word; 
  s[len] = 0;
  int length = stem(s,0,len-1)+1;
  s[length] = 0;
  word = s;
  len = length;
}

/*int main(int argc, char * argv[]){ 
  int t = 11;
  get_word_stem(*argv,t);
  printf("%s %i",*argv,t);
  return 1;
}*/